#ifndef _diffusion_client_control_h_
#define _diffusion_client_control_h_ 1

/*
 * Copyright © 2014, 2018 Push Technology Ltd., All Rights Reserved.
 *
 * Use is subject to license terms.
 *
 * NOTICE: All information contained herein is, and remains the
 * property of Push Technology. The intellectual and technical
 * concepts contained herein are proprietary to Push Technology and
 * may be covered by U.S. and Foreign Patents, patents in process, and
 * are protected by trade secret or copyright law.
 */

/**
 * @file client-control.h
 *
 * Client control feature.
 */

#include "session.h"

#define PROPERTIES_SELECTOR_ALL_FIXED_PROPERTIES "*F"
#define PROPERTIES_SELECTOR_ALL_USER_PROPERTIES "*U"

typedef enum {
        /// Indicates a session has been opened (e.g. new client
        /// connection).
        SESSION_PROPERTIES_EVENT_TYPE_OPEN = 0,
        /// Indicates that a session has updated properties.
        SESSION_PROPERTIES_EVENT_TYPE_UPDATE = 1,
        /// Indicates that a session has been closed.
        SESSION_PROPERTIES_EVENT_TYPE_CLOSE = 2,
        /// Indicates that a session has been disconnected
        SESSION_PROPERTIES_EVENT_TYPE_DISCONNECTED = 3
} SESSION_PROPERTIES_EVENT_TYPE_T;

typedef enum {
        /// The session properties have been updated.
        SESSION_PROPERTIES_UPDATE_TYPE_UPDATED = 0,
        /// The session properties have been updated due to
        /// reconnection with the server.
        SESSION_PROPERTIES_UPDATE_TYPE_RECONNECTED = 1,
        /// The session properties have been updated due to the
        /// session failing over to a different host.
        SESSION_PROPERTIES_UPDATE_TYPE_FAILED_OVER = 2,
        /// Session has disconnected
        SESSION_PROPERTIES_UPDATE_TYPE_DISCONNECTED = 3
} SESSION_PROPERTIES_UPDATE_TYPE_T;

typedef struct session_properties_event_s {
        /// The reason why a property event was triggered.
        SESSION_PROPERTIES_EVENT_TYPE_T type;
        /// If the session was closed, this indicates the reason for
        /// closure.
        SESSION_CLOSE_REASON_T close_reason;
        /// If the session properties were updated, this indicates the
        /// reason for the change.
        SESSION_PROPERTIES_UPDATE_TYPE_T update_type;
        /// The session ID of the client session.
        SESSION_ID_T session_id;
        /// A key/value map of properties of the client, as requested
        /// when registering the session properties listener.
        HASH_T *properties;
} SESSION_PROPERTIES_EVENT_T;

typedef int (*on_session_properties_registered_cb)(SESSION_T *session, void *context);

/// Callbacks for session open/close and property updates have this
/// function signature.
typedef int (*on_session_properties_event_cb)(SESSION_T *session,
                                              const SESSION_PROPERTIES_EVENT_T *event,
                                              void *context);

#define SESSION_PROPERTIES_REGISTRATION_HANDLERS                        \
        /** Callback fired when the listener has been registered. */    \
        on_session_properties_registered_cb on_registered;              \
        /** Error callback if an error occurs during listener registration.  */ \
        ERROR_HANDLER_T on_registration_error;                          \
        /** Standard discard handler. */                                \
        DISCARD_HANDLER_T on_discard;

#define SESSION_PROPERTIES_EVENT_HANDLERS                               \
        /** Callback fired when a new client session is opened. */      \
        on_session_properties_event_cb on_session_open;                 \
        /** Callback fired when an existing client session closes. */   \
        on_session_properties_event_cb on_session_close;                \
        /** Callback fired when a client's session detail are updated. */ \
        on_session_properties_event_cb on_session_update;               \
        /** Callback fired when a client was disconnected. */           \
        on_session_properties_event_cb on_session_disconnect;           \
        /** Error callback if an error occurs with a session update.  */ \
        ERROR_HANDLER_T on_session_error;

typedef struct session_properties_registration_handlers_s {
        SESSION_PROPERTIES_REGISTRATION_HANDLERS
        SESSION_PROPERTIES_EVENT_HANDLERS
} SESSION_PROPERTIES_REGISTRATION_HANDLERS_T;

typedef struct session_properties_registration_params_s {
        SESSION_PROPERTIES_REGISTRATION_HANDLERS
        SESSION_PROPERTIES_EVENT_HANDLERS
        /// A set containing properties of clients to be reported back
        /// to the listener. The properties are NULL-terminated
        /// strings, or the constants PROPERTIES_SELECTOR_ALL_FIXED_PROPERTIES or
        /// PROPERTIES_SELECTOR_ALL_USER_PROPERTIES.
        SET_T *required_properties;
        /// User-supplied context to return to callbacks.
        void *context;
} SESSION_PROPERTIES_REGISTRATION_PARAMS_T;

typedef struct svc_get_session_properties_response_s {
        /// Returned properties for the session.
        HASH_T *properties;
} SVC_GET_SESSION_PROPERTIES_RESPONSE_T;

typedef int (*get_session_properties_cb)(SESSION_T *session,
                                         const SVC_GET_SESSION_PROPERTIES_RESPONSE_T *response,
                                         void *context);

#define GET_SESSION_PROPERTIES_HANDLERS                             \
        /** Callback fired when session properties are received. */ \
        get_session_properties_cb on_session_properties;            \
        /** Error callback. */                                      \
        ERROR_HANDLER_T on_error;                                   \
        /** Discard callback. */                                    \
        DISCARD_HANDLER_T on_discard;

typedef struct get_session_properties_params_s {
        GET_SESSION_PROPERTIES_HANDLERS
        /// The session ID of the connected client whose properties we
        /// are requesting.
        SESSION_ID_T *session_id;
        /// A set containing properties of clients to be reported back
        /// to the listener. The properties are NULL-terminated
        /// strings, or the constants PROPERTIES_SELECTOR_ALL_FIXED_PROPERTIES or
        /// PROPERTIES_SELECTOR_ALL_USER_PROPERTIES.
        SET_T *required_properties;
        /// User-supplied context to return to callbacks.
        void *context;
} GET_SESSION_PROPERTIES_PARAMS_T;

/**
 * @ingroup PublicAPI_ClientControl
 *
 * @brief Register a listener for changes to other clients' properties.
 *
 * @param session       The current session. If NULL, this function returns immediately.
 * @param params        Parameter structure describing which properties should
 *                      be reported, and the callbacks to be invoked.
 */
void session_properties_listener_register(SESSION_T *session, const SESSION_PROPERTIES_REGISTRATION_PARAMS_T params);

/**
 * @ingroup PublicAPI_ClientControl
 *
 * @brief Request properties of a connected client.
 *
 * @param session       The current session. If NULL, this function returns immediately.
 * @param params        Parameter structure describing which client and
 *                      properties we're interested in, and the callbacks
 *                      to be invoked.
 */
void get_session_properties(SESSION_T *session, const GET_SESSION_PROPERTIES_PARAMS_T params);

#endif
